// ESA (C) 2000-2018 
//
// This file is part of ESA's XMM-Newton Scientific Analysis System (SAS).
//
//    SAS is free software: you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation, either version 3 of the License, or
//    (at your option) any later version.
//
//    SAS is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with SAS.  If not, see <http://www.gnu.org/licenses/>.
//
// Specgroup.cc
//
// Authors:  Richard Saxton (VEGA/ESA)
//           ESA XMM SOC
//
// Description: Task to group the channels of a spectrum
//              from a user supplied formula.
//
// NB: There is no reason why this should be XMM specific so endeavour to 
//     avoid any XMM specifics with keywords etc.
//
#include "CalServer.h"
#include "Dal.h"
#include "DatafileAttributes.h" 

#include "Specgroup.h"
#include "WishInterpreter.h"

#include "Tracer.h"
#include "errstr.h"
#include "param.h"
#include "slatec.h"
#include <unistd.h> 

/* Method: Specgroup() - Constructor */
Specgroup::Specgroup()
{
///  PROC;

  /* Initiate a variable to hold the spectral information */
    WishInterpreter::Spectrum srcspec;

  /* Get the parameters from this command line */
    srcspec.srcfile = stringParameter("spectrumset");

  /* Create the output spectrum and pointers */
    DataSet * outset=0;
    DataSet * cloneset=0;

   // Copy the input file into a new output file
    string outfile;
    if (! booleanParameter("overwrite")) {

       outfile = stringParameter("groupedset");
       outset = dataSetServer->clone( srcspec.srcfile, outfile, Dal::Modify);

    } else { // Set the input spectrum as the output spectrum

       outset = dataSetServer->open( srcspec.srcfile, DataSet::Modify);

      // Need to make a clone of this file which can be read but not
      // altered in other classes. Otherwise, we can't overwrite the input.
       srand(getpid());
       string clonefile = "temp_specgrp" + toString(rand()) + ".del";
       cloneset = dataSetServer->clone( srcspec.srcfile, clonefile,
					Dal::Temp);
       srcspec.srcfile = clonefile;
       outfile=srcspec.srcfile;
    }

   // Open the input/output spectrum table
    Table * spectrumtable = outset->table("SPECTRUM");

   // Get pointer to the channel column to get RGS TC keywords 
    Column *chancol = spectrumtable->column("CHANNEL");

   // Get the conversion factors from PI channel to channel in spectrum
   // EPIC use the SPECDELT, SPECPIX, SPECVAL keywords
   // RGS use the TCRPX1, TCRVL1, TCDLT1 keywords
    srcspec.gotPIconv = true;

    if (spectrumtable->hasAttribute("SPECDELT"))
       srcspec.specdelt=spectrumtable->attribute("SPECDELT")->asReal();       
    //else if (chancol->hasAttribute("TCDLT"))
      // srcspec.specdelt=chancol->attribute("TCDLT")->asReal();
    else 
       srcspec.specdelt=1;
    
    if (spectrumtable->hasAttribute("SPECPIX"))
       srcspec.specpix=spectrumtable->attribute("SPECPIX")->asReal();       
    //else if (chancol->hasAttribute("TCRPX"))
      // srcspec.specpix=chancol->attribute("TCRPX")->asReal();
    else 
       srcspec.specpix=1;
    
    if (spectrumtable->hasAttribute("SPECVAL"))
       srcspec.specval=spectrumtable->attribute("SPECVAL")->asReal();       
    //else if (chancol->hasAttribute("TCRVL")){
      // srcspec.specval=chancol->attribute("TCRVL")->asReal();
    else 
       srcspec.specval=1;  // Set so that Channel 0 = PI 1  ....

    errstr<<"Spec conversion: "<<srcspec.gotPIconv<<" "<<srcspec.specpix
        <<" "<<srcspec.specval<<" "<<srcspec.specdelt
        << Sas::message(Msg::App, Msg::Verbose);

   // Find if RGS Lambda or Beta spectrum 
    if (chancol->hasAttribute("TCUNI"))
       srcspec.rgsunits=chancol->attribute("TCUNI")->asString();

   // Get the counts - works for int*16 and int*32
    vector<int> scounts_in;
    CalOalUtility::fillVector(spectrumtable, "COUNTS", scounts_in);

    //Column *cntscol = spectrumtable->column("COUNTS");
    //ColumnData *cntsdata = cntscol->data();
    //int32 *scounts_in = cntsdata->int32Data();

   // Get the channels
    vector<int> channel_in;
    CalOalUtility::fillVector(spectrumtable, "CHANNEL", channel_in);

    //Column *chancol = spectrumtable->column("CHANNEL");
    //ColumnData *chandata = chancol->data();
    //int16 *channel_in = chandata->int16Data();

   // Read the background area
    double sbackscal=getBackscale(spectrumtable);

    errstr<<"Source - Backscale: "<<sbackscal
        << Sas::message(Msg::App, Msg::Verbose);

    unsigned npi=spectrumtable->rows();

  // Set up the source and background counts arrays as vectors
    for (unsigned i=0; i<npi; ++i) {
                srcspec.counts.push_back(scounts_in[i]);
                srcspec.channels.push_back(channel_in[i]);
                srcspec.bcounts.push_back(0);
    }

  /* Set the area normalisation to 1 by default */
    srcspec.areanorm=1.0;

  /* Read in and normalise the background counts if file has been given */
    if (booleanParameter("withbgdset")) {
       string bgdfile = stringParameter("backgndset");
       DataSet * backgndset = dataSetServer->open( bgdfile, DataSet::Read );
       Table * backgndtable = backgndset->table("SPECTRUM");
       
      // Handle the case of background COUNTS but not RATE
       if (!backgndtable->hasColumn("COUNTS")) 
          errstr<<"No COUNTS column found in the background file. NB: count rates are not currently handled."<<Sas::error("NOCOUNTS");

       vector<int> bcounts_in;
       CalOalUtility::fillVector(backgndtable, "COUNTS", bcounts_in);

       //Column *bcntscol = backgndtable->column("COUNTS");
       //ColumnData *bcntsdata = bcntscol->data();
       //int32 *bcounts_in = bcntsdata->int32Data();

      // Ensure that source and background have the same binning
       if (backgndtable->rows() != npi) 
          errstr<<"Source and background spectra do not have the same size"
                <<Sas::error("DIFFSIZES");

       double bbackscal=getBackscale(backgndtable);

       errstr<<"Backgnd - Backscale: "<<bbackscal
        << Sas::message(Msg::App, Msg::Verbose);


      // Get the relative exposure time
       double relexp = 1.0;

       if (spectrumtable->hasAttribute("EXPOSURE") &&
           backgndtable->hasAttribute("EXPOSURE")) {

          relexp = spectrumtable->attribute("EXPOSURE")->asReal() /
                   backgndtable->attribute("EXPOSURE")->asReal();
       }

      /* Set the area normalisation */
       srcspec.areanorm = sbackscal / bbackscal * relexp;

      // Set the background counts
       for (unsigned i=0; i<npi; ++i)
          srcspec.bcounts[i] = bcounts_in[i];

      // Clean
       //delete bcntsdata;
       dataSetServer->close(backgndset);
    }

  // Clean up
    //delete cntsdata;
    //delete chandata;

  /* Read and interpret the user's grouping wishes */
    WishInterpreter *wi = new WishInterpreter();

    if (booleanParameter("withtemplate")) 
       wi->copyFromTemplate(srcspec);
    else 
       wi->interpretWish(srcspec);

  /* Write the grouping information into the output file */
    writeGroups(spectrumtable, wi->groups());

  /* Write the QUALITY flags into the output file */
    writeQuality(spectrumtable, wi->quality());

  /* Write associated filenames into the header */
    writeAssociated(spectrumtable);

  /* Close down */
    dataSetServer->close(outset);
    if (cloneset) dataSetServer->close(cloneset);

  /* Tell the user what has happened */
    errstr<<"Prepared spectrum "<<outfile<<" with "<<wi->groups().size()<<" groups"
          << Sas::message(Msg::App, Msg::Sparse);

} // end

/* Destructor */
Specgroup::~Specgroup()
{
  PROC;

  ENDPROC;
}

void Specgroup::writeGroups(Table *spectrumtable, 
                               vector<WishInterpreter::Group> groups)
{
  /* Create a GROUPING column in the output table if it doesn't already exist */
     Column * GCol;
     if (! spectrumtable->hasColumn("GROUPING"))
        GCol = spectrumtable->addColumn("GROUPING", Column::Int16, 
                                        "Group constraint", "");
     else
        GCol = spectrumtable->column("GROUPING");

  /* Write the group data */
     ColumnData * GColData   = GCol->data();
     int16      * GrpData   = GColData->int16Data();

  /* Check that the vector sizes agree */
     // assert(GrpData->size() = cstart.size());

  /* Write the Grouping data */
     for (unsigned i=0; i<spectrumtable->rows(); ++i)
         GrpData[i]=1;  // Set all channels as single groups by default

    // Loop over each group
     for (unsigned i=0; i<groups.size(); i++) {

       // Set the initial member of the group to +1
        GrpData[groups[i].first]=1;

       // Set subsequent members to -1
        for (int j=groups[i].first+1; j<=groups[i].second; j++)
           GrpData[j]=-1;
     }

  /* Close up */
     delete GColData;

}
/* Method: writeQuality
   Purpose: writes the QUALITY array into the output spectrum
*/
void Specgroup::writeQuality(Table *spectrumtable, vector<short int> inQual)       
{
  /* Create a QUALITY column in the output table if it doesn't already exist */
     Column * QCol;
     if (! spectrumtable->hasColumn("QUALITY"))
        QCol = spectrumtable->addColumn("QUALITY", Column::Int16,
                         "Quality flag of this channel (0=good)", "");
     else
        QCol = spectrumtable->column("QUALITY");

  /* Write the group data */
     ColumnData * QColData   = QCol->data();
     int16      * QualData   = QColData->int16Data();

  /* Check that the vector sizes agree */
     //assert(QualData.size() = inQual.size());

  /* Write the quality data */
     for (unsigned i = 0; i<inQual.size(); i++)
         QualData[i] = inQual[i];

  /* Close up */
     delete QColData;
}

/* Method: writeAssociated
   Purpose: writes the ANCRFILE, RESPFILE, BACKFILE keywords
*/
void Specgroup::writeAssociated(Table *spectrumtable)
{

  /* Remove any GROUPING keyword - used to say that the spectrum 
     is not grouped */
    if (spectrumtable->hasAttribute("GROUPING"))
       spectrumtable->deleteAttribute("GROUPING");

  /* Should we update the keywords */
    if (booleanParameter("addfilenames")) {
 
      // Use the order - 1) set as parameter, 2) original value, 3) NONE
 
      // Write ANCRFILE keyword
       if (booleanParameter("witharfset")) 
          * spectrumtable->addAttribute("ANCRFILE", 
                   "ancillary response") = stringParameter("arfset");
       else if (!spectrumtable->hasAttribute("ANCRFILE")) 
          * spectrumtable->addAttribute("ANCRFILE",
                   "ancillary response") = "NONE";
          
      // Write RESPFILE keyword
       if (booleanParameter("withrmfset")) 
          * spectrumtable->addAttribute("RESPFILE", 
                   "redistribution matrix") = stringParameter("rmfset");
       else if (!spectrumtable->hasAttribute("RESPFILE")) 
          * spectrumtable->addAttribute("RESPFILE",
                   "redistribution matrix") = "NONE";

      // Write BACKFILE keyword
       if (booleanParameter("withbgdset")) 
          * spectrumtable->addAttribute("BACKFILE",
                   "Background FITS file") = stringParameter("backgndset");
       else if (!spectrumtable->hasAttribute("BACKFILE"))
          * spectrumtable->addAttribute("BACKFILE",
                   "Background FITS file") = "NONE";
    }
}

/* Method: getBackscale
   Purpose: Finds the BACKSCAL value from either the spectrum keyword
            or the maximum of a column in the case of RGS
*/
double Specgroup::getBackscale(Table *spectrumtable)
{
    double backscale=1.0; // Set default value if no information available

    if (spectrumtable->hasAttribute("BACKSCAL"))
       backscale=spectrumtable->attribute("BACKSCAL")->asReal();

    else if (spectrumtable->hasColumn("BACKSCAL")) 
       {
          vector<double> backs_in;
          CalOalUtility::fillVector(spectrumtable, "BACKSCAL", backs_in);       

         // Find the maximum of the backscale values
          double bmax=0.0;
          for (unsigned i=0; i<backs_in.size(); ++i) {
              if (backs_in[i] > bmax) 
                 bmax=backs_in[i];
          }
          backscale=bmax;
        }

    return backscale;
}
     

