// ESA (C) 2000-2018 
//
// This file is part of ESA's XMM-Newton Scientific Analysis System (SAS).
//
//    SAS is free software: you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation, either version 3 of the License, or
//    (at your option) any later version.
//
//    SAS is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with SAS.  If not, see <http://www.gnu.org/licenses/>.
//
// Module: WishInterpreter
// Author: Richard Saxton 
//
// Description: Main class for calculating  a source fluxer profile
//
// Modifications: See ChangeLog
//

#ifndef WishInterpreter_H
#define WishInterpreter_H

/* General classes */
#include <iostream>

/* SAS-specific classes */
#include "Tracer.h"
#include "Dal.h"

/* Local Structure */

/*
struct Response {
	vector<vector<float> > matrix;
	vector<double> elo,
		       ehi,
		       eblo,
		       ebhi;
};
*/

/****************************************************
 *
 * class WishInterpreter
 *
 ***************************************************/

class WishInterpreter
{
public:

  /* Constructor */
  WishInterpreter();

  /* Destructor */
  ~WishInterpreter();

  /* Member fields */

  typedef pair<int, int> Group;   // Define the structure of a Group

  struct Spectrum {
        string srcfile;   // Name of the input spectrum
        vector<int> channels;
        vector<int> counts;
        vector<int> bcounts;
        double areanorm;

      // Conversion from PI chans to spectral channel
        bool gotPIconv;
        double specdelt, specpix, specval;
    
      // RGS spectral units which hive the type
        string rgsunits;
  };

  /* Methods */
  void interpretWish(Spectrum srcspec);

    /* Purpose:
		Sets the start and end arrays for the channels based on 
		the users desires.

       Input:
                srcspec:  A structure containing :

   		channels: The array of the channel numbers
		counts:   The array of the spectral counts
		bcounts:  The array of background counts
                areanorm: The BACKSCAL normalistion of geometrical areas
                conv:     Variables to convert from channel to PI
    */

  void copyFromTemplate(Spectrum srcspec);

    /* Purpose:
                Copies groups from a template input spectrum

       Input:
                srcspec:  A structure containing :
    */

  void setRanges(vector<int> channels, string & rangeString);

    /* Purpose:
		Sets the ranges that the users desires.

       Input:
		A string containing range information
		e.g. [10-50,73-100]
    */

  bool checkRangeString(string & rangeString);

    /* Purpose:
                Check that the ranges format is ok

       Input:
                A string containing range information
                e.g. [10:50,73:100]
    */

  vector<Group> convertRangeString(vector<int> channels, string & rangeString);

    /* Purpose:
                Converts a text string containing ranges
                e.g. 1-10,11-15  into a set of groups

       Input:
                A string containing range information
                e.g. [10-50,73-100]
    */

  Group groupChanOrEnergy(vector<int> channels, string &startStr, string &endStr);

    /* Purpose:
                Makes a group from a range in channel or energy space

       Input:
                The (PI) channel number of each spectral bin
                A string containing the start and end of the range
                e.g. 10-50  or  0.2-2.0 (keV)
    */

  vector<Group> convertGroupVector(vector<int16> grpVector);

    /* Purpose:
                Converts a vector of group values into a set of groups

       Input:
                The grouping column from a spectrum
    */

  void updateUsed();

    /* Purpose:
		Sets a flag for each channel in a group to say that it
		has been used and can't be used again
    */

  void regularBins(int rbin_start, int rbin_end, int rbin_width);

    /* Purpose:
		Sets groups based on regular binning

       Input:
		The start, end and width of the bins
    */

  const vector<Group> getBadRangefromCCF(Spectrum srcspec);

    /* Purpose:
                Find the range of channels to be excluded from a
                Calibration file
    */

  void setBadChans(Spectrum srcspec);

    /* Purpose:
                To set a range of channels bad 
    */

  
  const Group findspecChanG(int startPI, int endPI, Spectrum srcspec);

    /* Purpose:
                To set a group from a start and end PI channel
    */

  int findspecChan(double inval, Spectrum srcspec);

    /* Purpose:
                Returns the spectral channel corresponding to a PI channel
		or an input wavelength or beta value
    */

  void statGroup(vector<int> counts, vector<int> bcounts, 
                     double areanorm, double minStat, const string function);

    /* Purpose:
                Groups channels when they reach a minimum signal-to-noise ratio

       Input:
                counts:   The array of the spectral counts
                bcounts:  The array of background counts
                areanorm: The ratio of source to background areas
                minSN:    The minimum acceptable S/N ratio
                function: Which statistic to use - options are:

                          a) withCounts
                          b) withminSN
                          c) withratioabovebgnd
    */

   void handleUngrouped();

    /* Purpose:
                Sets ungrouped channels depending on the 'lastbin' flag
    */

   void mergeGroups(Group newgroup);

    /* Purpose:
          Merges a set of ungrouped channels into an existing group
    */

   void mergeGroups(unsigned groupNumber);

    /* Purpose:
          Merges an existing group with its neighbour (if possible)
    */

   void resolveOversampling(float oversample);

    /* Purpose:
          Ensures that the a group isnt too finely binned
    */

   double energy_of_channel(int chan);

    /* Purpose:
          Find the energy of an input channel
    */

   int channel_of_energy(double energy);

    /* Purpose:
          Find the channel of an input energy
    */

   int bin_of_channel(vector<int> channels, int inchan);

    /* Purpose:
          Find the bin in the spectrum containing a given channel
    */

   void SetCALfromSpec(const string & specfile);

    /* Purpose:
          Set the CAL status from the instrument and observing mode keywords
    */


  /* Output methods */
   vector<Group> groups();   // Return the groups
   vector<short int> quality();   // Return the quality array
   

  private:

    vector<Group> _groups;  // First and last channel of each group

    string _rangeString;  // string containing ranges

    vector<bool> _UsedSoFar; // Keeps tally of the channels which 
                             // have been grouped
    vector<short int> _quality; // Quality of each channel

    string _units;         // Units for user input (energy or channel)
    vector<int> _echan;     // The channel number of each energy bound
    vector<double> _chan_energy;     // The central energy of each channel
    bool _high2Low;        // Whether to start grouping from the high energy end

    enum ChanQual { GOOD, BAD, DUBIOUS };

    bool _setbadrange;     // Whether bad channels are defined as a range

   // Define a structure to hold the spectral information
   /*
    struct Spectrum {
        vector<int> channels;
        vector<int> counts;
        vector<int> bcounts;
        double areanorm;
    };
   */
};

#endif

